implicit none
integer, parameter :: nq = 9, nrand = 2, nqq = nq*(nq+1)/2, mq = 2*nqq, nc = 2
integer                       :: nrep, ii, i, j,ij, zrep, mrep, nrank
logical                       :: pen
real(8)                       :: xx, yy, zz, ee, hh, lbar, lvar, ess, edet, gdet, pdet
real(8), dimension(nq,1)      :: eig, ebar, epop
real(8), dimension(nq,nq)     :: evar
real(8), dimension(nqq)       :: ww
real(8), dimension(3,0:nc)    :: xloss
real(8), dimension(nq,nq)     :: gg, tt, zige, zigg, zigp
real(8), dimension(mq,1)      :: sig, sbar, zbar, pop
real(8), dimension(mq,mq)     :: ss
real(8), dimension(8)         :: diff, dbar
real(8), external             :: entropy

! population values
open(1,file='popvalues',action='read')
do i = 1, mq
   read(1,*) pop(i,1) ! covariances
end do
ww = pop(:nqq,1)
print *,ww
call ginverse( ww, eig(:nq,1), edet, 1.d-12, nrank, nq, .false. )
ij = 0
do i = 1, nq
   do j = i, nq
      ij = ij + 1; zige(j,i) = ww(ij); zige(i,j) = ww(ij)
   end do
end do
ww = pop(nqq+1:,1)
call ginverse( ww, eig(:nq,1), gdet, 1.d-12, nrank, nq, .false. )
ij = 0
zigg= 0.d0
do i = 1, nq
   do j = i, nq
      ij = ij + 1; zigg(j,i) = ww(ij); zigg(i,j) = ww(ij)
   end do
end do
ww = pop(:nqq,1) + pop(nqq+1:,1)
call ginverse( ww, eig(:nq,1), pdet, 1.d-12, nrank, nq, .false. )
ij = 0
do i = 1, nq
   do j = i, nq
      ij = ij + 1; zigp(j,i) = ww(ij); zigp(i,j) = ww(ij)
   end do
end do

do i = 1, nq
   read(1,*) epop(i,1) ! can e-values
end do
close(1)

mrep = 9999
! unpenalized estimates
pen = .false.
call get_samples( 0, 8, 'BestAll0' )
! penalized ....
pen = .true.
call get_samples( 1, 11, 'BestAll8Can' )
call get_samples( 2, 11, 'BestAll8PAC' )

contains

subroutine get_samples( icase, ll, fname )

integer, intent(in)           :: icase, ll
character(len=ll), intent(in) :: fname
real(8), dimension(nq,nq)     :: vv, ghat, ehat, phat, ping
real(8)                       :: es, yy, pp
real(8), dimension(3)         :: loss
integer                       :: i, j, ij, nr

open(1,file = fname, action = 'read', status = 'old' )
nrep = 0; sbar = 0.d0; ss = 0.d0; ebar = 0.d0; evar = 0.d0
ess = 0.d0; lbar = 0.d0; lvar = 0.d0; loss = 0.d0

! read replicates
do
! .. unpenalized log L
  if( pen ) then
      read(1,*,iostat=ii) yy, i, ess, j, xx, pp; if( ii /= 0 ) exit ! log L
  else
      read(1,*,iostat=ii) xx; if( ii /= 0 ) exit ! log L
  end if
  lbar = lbar + xx; lvar = lvar + xx * xx
  read(1,*) sig(:,1) ! sige + sigg
  nrep = nrep + 1
  sbar = sbar + sig; ss = ss + matmul( sig, transpose(sig) )
  ij = 0
  do i = 1, nq
     do j = i, nq
        ij = ij +1; ehat(j,i) = sig(ij,1); ehat(i,j) = sig(ij,1)
        ghat(j,i) = sig(ij+nqq,1); ghat(i,j) = ghat(j,i)
     end do
  end do
  phat = ghat + ehat
! canonical eigenvalues
  call eig_pinvg(nq, 1.d-10, 1.d-8, es, eig(:,1), ghat, ehat, ping, vv, nr)
  ebar = ebar + eig; evar = evar + matmul( eig, transpose(eig) )

! loss
  loss(1) =  loss(1) + entropy( nq, gdet, ghat, zigg )
  loss(2) =  loss(2) + entropy( nq, edet, ehat, zige )
  loss(3) =  loss(3) + entropy( nq, pdet, phat, zigp )

end do
close(1)
ss = ( ss - matmul( sbar, transpose(sbar) )/dble(nrep) )/ dble(nrep-1)
sbar = sbar/dble(nrep)
evar = ( evar - matmul( ebar, transpose(ebar) )/dble(nrep) )/ dble(nrep-1)
ebar = ebar/dble(nrep)
lvar = ( lvar -  lbar * lbar /dble(nrep) )/ dble(nrep-1)
lbar = lbar/dble(nrep)
loss = loss/dble(nrep); xloss(:,icase) = loss
ij = 0
do i = 1, nq
   do j = i, nq
!      ij = ij + 1; print *,'e',i,j, sbar(ij,1), sqrt(ss(ij,ij))
   end do
end do
do i = 1, nq
   do j = i, nq
!      ij = ij + 1; print *,'g',i,j, sbar(ij,1), sqrt(ss(ij,ij))
   end do
end do
print *,fname, nrep
do i = 1, nq
   print '(i3,3f12.6)',i, epop(i,1), ebar(i,1), sqrt(evar(i,i))
end do
print *,'loss', loss
if( pen ) then
    do i = 1, 3
       loss(i) = 100.d0*( 1.d0 - xloss(i,icase)/xloss(i,0) )
    end do
    print *,'prial', loss
end if
print '(i3,3f12.3)',0, ess, lbar, sqrt(lvar)


return 
end subroutine get_samples

end

!=====================================================================
      subroutine eigem (sig, eig, zero, eigzer, es, vv, nq, ineg )
!=====================================================================

      implicit none
      real(8),intent(in)                              :: zero, eigzer
      integer, intent(in)                             :: nq
      real(8), intent(out)                            :: es
      real(8) , dimension(nq*(nq+1)/2), intent(in)    :: sig
      real(8) , dimension(nq), intent(out)            :: eig
      real(8) , dimension(nq,nq), intent(out)         :: vv
      integer, intent(out)                            :: ineg

      real(8) , dimension(nq*(nq+1)/2)                :: work
      real(8), dimension(nq,nq)                       :: ww
      integer, external                               :: ihmii, ihmssf
      integer :: n, nsweep, i, j, ij, ip, iq, ipp, iqq, ipq, ijp, ijq&
           &, max=800
      real(8) :: ezero, eps=1.d-6
      real(8) :: c, s, cc, ss, cs, csa, app, aqq, apq, theta, t, aijp&
           &, aijq, vp, vq, x

      ezero = zero; vv = 0.d0
      work = sig
      n=nq; nsweep=0

      ww=0.d0
      do i=1,n
      ww(i,i)=1.d0
      end do

!     CALCULATE SUM OF SQUARES OF OFF-DIAGONAL ELEMENTS
      do while (nsweep<max)

      SS=0.D0; IJ=0
      DO I=1,N-1
      DO J=I,N
         IJ=IJ+1; if(i==j)cycle
         SS=SS+work(ij)*work(ij)
      end do
      end do

!     TEST WHETHER SS ARE SUFFICIENTLY SMALL
      if(2.d0*ss<eps*eps)then
         ineg=0
         do i=1,n
         work(i)=work(ihmii(i,n))
         end do
!        sort : largest eigenvalues first
         do i=1,n
         x=-1.d6
         do j=1,n
         if(work(j)>x)then
            x=work(j); ij=j
         end if
         end do
         eig(i)=x
! ij=i; eig(i)=work(i) ! switch off re-ordering
         if(x < eigzer) ineg=ineg+1
         work(ij)=-1.d12; vv(:n,i)=ww(:n,ij)
         end do
         es = sum (eig(:nq) ); return
      end if

!     MATRIX NOT DIAGONAL YET, NEED MORE TRANSFORMATIONS
      IPP=-N
      IPQ=0
      DO IP=1,N-1
      IPP=IPP+N+1
      APP=work(IPP)
      IPQ=IPQ+1
      DO IQ=IP+1,N
      IPQ=IPQ+1
      APQ=work(IPQ)

      IF(abs(APQ)>ezero)THEN
!        APPLY P-Q ROTATION TO MATRIX 
         IQQ=IHMII(IQ,N)
         AQQ=work(IQQ)
         THETA=0.5D0*(AQQ-APP)/APQ
         IF(abs(THETA)<ezero)THEN
            T=1.D0
         ELSE
            X=DSQRT(1.D0+THETA*THETA)
            T=1.D0/(abs(THETA)+X)
            if(theta<0)t=-t
         END IF
         C=1.D0/DSQRT(1.D0+T*T)
         S=C*T
         CC=C*C
         CS=C*S
         SS=S*S
         CSA=2.D0*CS*APQ
         work(IPP)=CC*APP-CSA+SS*AQQ
         work(IPQ)=CS*(APP-AQQ)+(CC-SS)*APQ
         work(IQQ)=SS*APP+CSA+CC*AQQ
         APP=work(IPP)
         DO  J=1,IP-1
         IJP=IHMssf(IP,J,N); IJQ=IHMSSF(IQ,J,N)
         AIJP=work(IJP); AIJQ=work(IJQ)
         work(IJP)=C*AIJP-S*AIJQ; work(IJQ)=S*AIJP+C*AIJQ
         end do
         IJP=IPP
         DO J=IP+1,IQ-1
         IJP=IJP+1; IJQ=IHMssf(IQ,J,N)
         AIJP=work(IJP); AIJQ=work(IJQ)
         work(IJP)=C*AIJP-S*AIJQ
         work(IJQ)=S*AIJP+C*AIJQ
         end do
         IJQ=IQQ
         DO J=IQ+1,N
         IJP=IHMssf(IP,J,N);IJQ=IJQ+1
         AIJP=work(IJP); AIJQ=work(IJQ)
         work(IJP)=C*AIJP-S*AIJQ
         work(IJQ)=S*AIJP+C*AIJQ
         end do

!        accumulate transformations to form matrix of eigenvectors
         DO I=1,N
         VP=ww(I,IP); VQ=ww(I,IQ)
         ww(I,IP)=C*VP-S*VQ; ww(I,IQ)=S*VP+C*VQ
         end do
      end if

      end do ! iq
      IPP=IPP-IP
      end do !ip

      NSWEEP=NSWEEP+1
      end do ! do while

      print *,'eigenvalue decomposition'
      PRINT *,'MAX. NO. OF SWEEPS REACHED    =',MAX
      PRINT *,'ACCURACY TOLERANCE GIVEN      =',EPS, ss
      print *, work
      stop 'numerical problems'
  !    eefail = .true.!; pause

      END subroutine eigem

! ===============================
  INTEGER FUNCTION IHMSSF(I,J,N)
! ===============================

  implicit none
  integer, intent(in) :: i, j, n
  integer             :: i1, j1

  IF(I.LE.J)THEN
     I1=I-1; IHMSSF=N*I1-I*I1/2+J
  ELSE
     J1=J-1; IHMSSF=N*J1-J*J1/2+I
  END IF

  RETURN
  END function ihmssf

! ============================
  INTEGER FUNCTION IHMII(I,N)
! ============================

  implicit none
  integer, intent(in) :: i, n

  IHMII=N*(I-1)+I*(3-I)/2

  RETURN
  END function ihmii


! ============================================================================
  subroutine eig_pinvg( mfit, zero, eigzer, es, eig, ghat, ehat, ping, vv, nr)
! =============================================================================

  implicit none
  integer, intent(in)                       :: mfit
  real(8), intent(in)                       :: zero, eigzer
  real(8), dimension(mfit,mfit), intent(in) :: ghat, ehat
  real(8), dimension(mfit,mfit), intent(out):: ping
  integer, intent(out)                      :: nr
  real(8), dimension(mfit), intent(out)     :: eig
  real(8), dimension(mfit,mfit), intent(out):: vv
  real(8), intent(out)                      :: es
  integer                                   :: i, j, ij
  real(8), dimension(mfit,mfit)             :: hh, lmat
  real(8), dimension(mfit*(mfit+1)/2)       :: ww

  hh = ghat + ehat
  ij = 0
  do i = 1, mfit
     do j = i, mfit
        ij = ij+1; ww(ij) = hh(j,i)
     end do
  end do
  call eigem(ww, eig, zero, eigzer, es, vv, mfit, nr)
  hh = 0.d0
  do i = 1, mfit
     hh(i,i) = 1.d0/sqrt(eig(i))
  end do
  lmat = matmul( vv, hh ) ! matrix square root of Sig_P^{-1}
  hh = matmul( transpose(lmat), ghat); ping = matmul(hh, lmat ); hh = ping
  ij = 0
  do i = 1, mfit
     do j = i, mfit
        ij = ij+1; ww(ij) = hh(j,i)
     end do
  end do
  call eigem( ww, eig, zero, eigzer, es, vv, mfit, nr)

  return
  end subroutine eig_pinvg

! ===============================================
  real(8) function entropy( nq, hdet, hhat, hinv )
! ===============================================

  implicit none
  integer, intent(in)                   :: nq
  real(8), intent(in)                   :: hdet
  real(8), dimension(nq,nq), intent(in) :: hhat, hinv
  real(8), dimension(nq,nq)             :: hh
  real(8), dimension(nq)                :: eig
  real(8), dimension(nq*(nq+1)/2)       :: ww
  real(8)                               :: xx, det
  integer                               :: i, j, ij, nrank

  hh = matmul( hhat, hinv) 
  xx = -dble(nq)
  do i = 1, nq
     xx = xx + hh(i,i)
  end do
  ij = 0
  do i = 1, nq
     do j = i, nq
        ij = ij + 1; ww(ij) = hhat(j,i)
     end do
  end do
  call ginverse( ww, eig, det, 1.d-12, nrank, nq, .false. )
  xx = xx + hdet - det
  entropy = xx

  return
  end function entropy
! =========================================================
  subroutine ginverse( sig, eig, det, zero, nrank, n, opt )
! =========================================================

  implicit none
  integer, intent(in)                          :: n
  real(8), dimension(n*(n+1)/2), intent(inout) :: sig
  real(8), dimension(n), intent(out)           :: eig
  real(8), intent(in)                          :: zero
  logical, intent(in)                          :: opt
  real(8), intent(out)                         :: det
  integer, intent(out)                         :: nrank
  
  real(8), dimension(n,n)                      :: vv
  real(8)                                      :: eigzer, xx
  integer                                      :: i, j, k, ij, ineg

  eigzer = 1.05d0*zero; eig = 0.d0; vv =0.d0

! eigen-decomposition
  call eigem (sig, eig, zero, eigzer, xx, vv, n, ineg )
  nrank = n - ineg

! inverse of eigenvalues
  det = 0.d0
  do i = 1,nrank
     det = det + log( eig(i) )
     eig(i) = 1.d0/eig(i)
  end do
  if( opt ) then
      write(*,*)'"ginverse" : rank, det =', n, nrank,det
      write(*,*)'eigs',eig
  end if

! construct inverse
  ij = 0
  do i = 1, n
     do j = i, n
        ij = ij + 1
        xx = 0.d0
        do k = 1, nrank
           xx = xx + vv(j,k) * vv(i, k) * eig(k)
        end do
        sig(ij) = xx
     end do
  end do

  return
  end subroutine ginverse
